package sets;


public class SequentialSet<T> implements Set<T>
{
    // nodes at beginning and end
    protected Node<T> head, tail;

    // create empty set
    public SequentialSet() {
        head = newNode(Integer.MIN_VALUE);
        tail = newNode(Integer.MAX_VALUE);
        try {
            head.setNext(tail);
        } catch (UnsupportedOperationException e) {
            // kludge: so that argumentless constructor can be reused in LockFreeSet
        }
    }

    // if `node.key' == `curr.key' return false
    // else insert `node' between `pred' and `curr' and return true
    protected boolean rawAdd(Node<T> pred, Node<T> curr, Node<T> node) {
        if (curr.key() == node.key())
            return false;  // item already in set
        else { // insert node between pred and curr
            node.setNext(curr);
            pred.setNext(node);
            return true;
        }
    }

    public boolean add(T item) {
        // new node
        Node<T> node = newNode(item);
        Node<T> pred = null, curr = null; // dummy initialization
        // first node from head whose key >= `node.key'
        Position<T> where = find(head, node.key());
        pred = where.pred;
        curr = where.curr;
        // pred.key < node.key <= curr.key
        return rawAdd(pred, curr, node);
    }

    // if `node.key' < `key' return false
    // else remove `node' by skipping over it from `pred' and return true
    protected boolean rawRemove(Node<T> pred, Node<T> node, int key) {
        if (node.key() > key)
            return false;  // item not in set
        else { // remove node curr
            pred.setNext(node.next());
            return true;
        }
    }

    public boolean remove(T item) {
        int key = item.hashCode();
        Node<T> pred = null, curr = null; // dummy initialization
        // first node from head whose key >= `key'
        Position<T> where = find(head, key);
        pred = where.pred;
        curr = where.curr;
        // pred.key < key <= curr.key
        return rawRemove(pred, curr, key);
    }

    // if `node.key' > `key' return false
    // else return true
    protected boolean rawHas(Node<T> node, int key) {
        if (node.key() > key)
            return false;  // item not in set
        else
            return true;
    }

    public boolean has(T item) {
        int key = item.hashCode();
        Node<T> pred = null, curr = null; // dummy initialization
        // first node from head whose key >= `key'
        Position<T> where = find(head, key);
        pred = where.pred;
        curr = where.curr;
        // pred.key < key <= curr.key
        return rawHas(curr, key);
    }

    // first position from `start' such that curr.key >= key
    protected Position<T> find(Node<T> start, int key) {
        Node<T> pred, curr; // predecessor and current node in iteration
        curr = start;      // from `start' node
        do { // move to next node
            pred = curr;
            curr = curr.next();
        } while (curr.key() < key);  // until curr.key >= key
        return new Position<T>(pred, curr);
    }

    // new Node<T> of appropriate subtype with given item
    protected Node<T> newNode(T item) {
        return new SequentialNode<>(item);
    }

    // new Node<T> of appropriate subtype with given key
    protected Node<T> newNode(int key) {
        return new SequentialNode<>(key);
    }
}
